(***************************************************************)
(* Coq code for PCAP                                           *)
(*                                                             *)
(* Map Implementation                                          *)
(*    A -> B : Type                                          *)
(*    Function                                                 *)
(***************************************************************)

(* $Id: FMapT.v,v 1.4 2006/07/15 02:56:41 ssg05 Exp $ *)

Require Import LIN_ORDER.
Require Import MAPT.

Set Implicit Arguments.
Unset Strict Implicit.

Module FuncMapT (ADef : LIN_ORDER) (BDef : ONE_TYPE) <: MAPDEF.

  Module ALib := LinLib ADef.

Definition A := ADef.A.

Definition beq := ADef.beq.

Lemma beq_true_eq : forall a a' : A, 
  beq a a' = true -> a = a'.
Proof.
  apply ADef.beq_true_eq.
Qed.

Lemma beq_false_neq : forall a a' : A,
    beq a a' = false -> a <> a'.
Proof.
  apply ADef.beq_false_neq.
Qed.

Lemma eq_beq_true : forall a a' : A, 
    a = a' -> beq a a' = true.
Proof.
  apply ADef.eq_beq_true.
Qed.

Lemma neq_beq_false : forall a a' : A, 
    a <> a' -> beq a a' = false.
Proof.
  apply ADef.neq_beq_false.
Qed.

Lemma eq_dec : forall a a' : A, {a=a'}+{a<>a'}.
Proof.
  apply ADef.eq_dec.
Qed.

Definition B := BDef.B.

(***********************)
(** map                *)
(***********************)

Definition map := A -> optionT B.

Definition emp_map : map := fun _ => NoneT.

Definition sig_map (a' : A) (b : B) : map := 
  fun a : A =>
    match (beq a a') with
      | true => SomeT b
      | false => NoneT
    end.

(***********************)
(** set                *)
(** get                *)
(***********************)


Definition get (m : map) (a : A) : optionT B := m a.

Lemma get_dec : forall (m : map) (a : A),
  {exists b, get m a = SomeT b} 
  + {get m a = NoneT}.
Proof.
  intros m a; unfold get.
  destruct (m a).
  left; exists b; trivial.
  right; trivial.
Qed.

Definition set (m : map) (a : A) (b : B) :=
  fun a' =>
    match (beq a a') with
      | true => SomeT b
      | false => m a'
    end.

Definition upd (m : map) (a : A) (b : B) : optionT map :=
  match (m a) with
    | SomeT b' => SomeT (fun a' => match (beq a a') with
                                     | true => SomeT b
                                     | false => m a'
                                   end)
    | NoneT => NoneT
  end.

Lemma set_sem : forall m a a' b, 
  match (beq a a') with
    | true => get (set m a b) a = SomeT b
    | false => get (set m a b) a' = get m a'
  end.
Proof.
  intros m a a' b.
  unfold set, get, beq.
  ALib.simplbeq.
  destruct (eq_dec a a'). 
  subst a';  ALib.simplbeq; trivial. 
  ALib.simplbeq; simpl; auto.
Qed.

Lemma upd_sem : forall m a b, 
  match get_dec m a with 
    | left _ => upd m a b = SomeT (set m a b)
    | right _ => upd m a b = NoneT
  end.
Proof.
  intros.
  destruct (get_dec m a) as [[b' Hin] | Hnin].
  unfold upd, set, get in * .
  rewrite Hin; simpl.
  trivial.
  unfold upd, set, get in * .
  rewrite Hnin; simpl.
  trivial.
Qed.

Lemma emp_sem : forall a : A, get emp_map a = NoneT.
Proof.
  intros a; unfold get, emp_map; trivial.
Qed.

Lemma sig_sem : forall (a a' : A) (b : B),
  get (sig_map a' b) a = 
  match (beq a a') with
    | true => SomeT b
    | false => NoneT
  end.
Proof.
  intros; unfold get, sig_map; trivial.
Qed.

(************************)
(** merge               *)
(************************)

Definition merge (m1 m2 : map) a :=
  match m1 a with
  | NoneT => m2 a
  | SomeT b => SomeT b
  end.

Lemma merge_sem : forall (m1 m2 : map) (a : A),
  get (merge m1 m2) a 
  = match get m1 a, get m2 a with
      | SomeT b1, SomeT b2 => SomeT b1
      | SomeT b1, NoneT => SomeT b1
      | NoneT, SomeT b2 => SomeT b2
      | NoneT, NoneT => NoneT
    end.
Proof.
  unfold get, merge; intros m1 m2 a.
  destruct (m1 a); destruct (m2 a); trivial.
Qed.

End FuncMapT.

Unset Implicit Arguments.
