(***************************************************************)
(* Coq code for PCAP                                           *)
(*                                                             *)
(* Tiny Machine                                                *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: machine.v,v 1.5 2006/07/26 02:39:54 ssg05 Exp $ *)

Require Export codeheap.
Require Export heap.
Require Export reg.
Require Export regfile.
Require Export label.
Require Export instr.
Require Export word.
Require Export natpm.

Notation Local "'do' x <-- o ; f" :=
  (match o with Some y => (fun x => f) y | None => None end)
  (at level 200, x ident) : tm_scope.

Notation Local "'ret' x" := (Some x) 
  (at level 200) : tm_scope.

Notation pcreg := (label) (only parsing).

Notation "'++' pc" := (pc + 1) 
  (at level 0, only parsing) : tm_scope.

Inductive state : Set :=
  | st_con : heap -> regfile -> pcreg -> state.

Notation " <. h , rf , pc .> " := (st_con h rf pc)
  (at level 0, rf at level 40, pc at level 50) : tm_scope.

Open Local Scope tm_scope.

Notation gR := (Regfile.get).

Notation sR := (Regfile.set).

Notation gH := (Heap.get).

Notation sH := (Heap.set).

Notation S_R := 
  (fun (s : state) =>
  match s with 
    | <.h, rf, pc.> => rf
  end).

Notation S_H := 
  (fun (s : state) =>
  match s with 
    | <.h, rf, pc.> => h
  end).

Notation S_pc := 
  (fun (s : state) =>
  match s with 
    | <.h, rf, pc.> => pc
  end).

Notation gSR :=
  (fun (s : state) (r : reg) =>
    match s with
      | <.h, rf, pc.> => 
        Regfile.get rf r
    end).

Notation gSH :=
  (fun (s : state) (l : label) =>
    match s with
      | <.h, rf, pc.> => 
        Heap.get h l
    end).

Notation sSR :=
  (fun (s : state) (r : reg) (w : word) =>
    match s with 
      | <.h, rf, pc.> =>
      <.h, Regfile.set rf r w, pc.>
    end).

Notation sSH :=
  (fun (s : state) (l : label) (w : word) =>
    match s with
      | <.h, rf, pc.> =>
      <.Heap.set h l w, rf, pc.>
    end).

Definition rget (s : state) (r : reg) : word :=
  match s with
    | <.h, rf, pc.> =>
    Regfile.get rf r
  end.

Definition rset (s : state) (r : reg) (w : word) : state :=
  match s with 
    | <.h, rf, pc.> => <.h, Regfile.set rf r w, pc.>
  end.

Definition hget (s : state) (l : label) : option word :=
  match s with 
    | <.h, rf, pc.> => Heap.get h l
  end.

Definition hset (s : state) (l : label) (w : word) : state :=
  match s with
    | <.h, rf, pc.> => <.Heap.set h l w, rf, pc.>
  end.

Definition pcinc (s : state) : state :=
  match s with
    | <.h, rf, pc.> => <.h, rf, ++pc.>
  end.

Definition pcset (s : state) (f : label) : state :=
  match s with 
    | <.h, rf, pc.> => <.h, rf, f.>
  end.

Definition r_s (s : state) := 
  match s with 
    | <.h, rf, pc.> => rf
  end.

Definition h_s (s : state) :=
  match s with 
    | <.h, rf, pc.> => h
  end.

Definition pc_s (s : state) :=
  match s with 
    | <.h, rf, pc.> => pc
  end.

Definition machine : Set := (codeheap * state) % type.

Implicit Types P M : machine.

Definition nextS (i : instr) (s : state) : option state :=
  match i with
    | addu rd rs rt =>
      match s with
        | <.h, rf, pc.> =>
          do sum <-- Some (gR rf rs + gR rf rt) ;
            do rf' <-- Some (sR rf rd sum) ;
              ret <.h, rf', ++pc.>
      end
    | addiu rd rs w =>
      match s with
        | <.h, rf, pc.> =>
          do sum <-- Some (gR rf rs + w) ;
            do rf' <-- Some (sR rf rd sum) ;
              ret <.h, rf', ++pc.>
      end
    | subu rd rs rt =>
      match s with
        | <.h, rf, pc.> =>
          do sum <-- Some (gR rf rs - gR rf rt) ;
            do rf' <-- Some (sR rf rd sum) ;
              ret <.h, rf', ++pc.>
      end
    | subiu rd rs w =>
      match s with
        | <.h, rf, pc.> =>
          do sum <-- Some (gR rf rs - w) ;
            do rf' <-- Some (sR rf rd sum) ;
              ret <.h, rf', ++pc.>
      end
    | lw rt i rs =>
      match s with
        | <.h, rf, pc.>  =>
          do data <-- gH h (gR rf rs + i) ;
            do rf' <-- Some (sR rf rt data);
              ret <.h, rf', ++pc.>
      end
    | sw rt i rs =>
      match s with
        | <.h, rf, pc.>  =>
          do data <-- Some (gR rf rt);
            do origdata <-- gH h (gR rf rs + i);
              do h' <-- Some (sH h (gR rf rs + i) data);
                ret <.h', rf, ++pc.>
      end
    | blt rs rt f =>
      match s with
        | <.h, rf, pc.>  =>
          if blt_nat (gR rf rs) (gR rf rt)
            then
              ret <.h, rf, f.>
            else
              ret <.h, rf, ++pc.>
      end
    | bgtz rs f =>
      match s with
        | <.h, rf, pc.>  =>
          if blt_nat O (gR rf rs)
            then
              ret <.h, rf, f.>
            else
              ret <.h, rf, ++pc.>
      end
    | j f =>
      match s with
        | <.h, rf, pc.>  =>
          ret <.h, rf, f.>
      end
    | jr r =>
      match s with
        | <.h, rf, pc.> =>
          do f <-- Some (gR rf r) ;
            ret <.h, rf, f.>
      end
  end.

(* Yet another next-state function, whose 
  return-term is smaller than the above one
  for efficiency consideration *)
Definition nextS2 (i : instr) (s : state) : option state :=
  match i with
    | addu rd rs rt => 
      Some (pcinc (rset s rd (rget s rs + rget s rt)))
    | addiu rd rs w => 
      Some (pcinc (rset s rd (rget s rs + w)))
    | subu rd rs rt => 
      Some (pcinc (rset s rd (rget s rs - rget s rt)))
    | subiu rd rs w => 
      Some (pcinc (rset s rd (rget s rs - w)))
    | lw rt w rs => 
      match hget s (rget s rs + w) with
        | Some v => Some (pcinc (rset s rt v))
        | None => None 
      end
    | sw rt w rs => 
      match (hget s (rget s rs + w)) with
        | Some _ => 
            Some (pcinc (hset s (rget s rs + w) (rget s rt)))
        | None => None
      end
    | blt rs rt f => 
      match blt_nat (rget s rs) (rget s rt) with
        | true => Some (pcset s f)
        | false => Some (pcinc s)
      end
    | bgtz rs f => 
      match blt_nat O (rget s rs) with
        | true => Some (pcset s f)
        | false => Some (pcinc s)
      end
    | j f => Some (pcset s f)
    | jr r => Some (pcset s (rget s r))
  end.


Definition step (P : machine) : option machine :=
  match P with
    | (ch, s) => 
      match s with 
        | <.h, rf, pc.> =>
          do i <-- codeheap.get ch pc ;
            do s' <-- nextS i s ;
              ret (ch, s')
      end
  end.

Fixpoint stepn (n : nat) (p : machine) {struct n} : option machine :=
  match n with
    | O => Some p
    | S n' =>
      match step p with
	| Some p' => stepn n' p'
	| None => None
      end
  end.

Notation "P ~~> P'" := (step P = Some P') 
  (at level 5) : tm_scope.

Notation optp :=
  (fun f => fun opt:option _ => 
    match opt with
      | Some t => f t
      | None => False
    end).

Notation optg :=
  (fun f => fun opt:option _ =>
    fun t' =>
      match opt with
        | Some t => f t t'
        | None => False
      end).

Notation " i ; I " := (iseq i I) 
  (at level 60, right associativity) : tm_scope.

Notation " i ;; " := (ione i)
  (at level 60) : tm_scope.

Ltac unfoldgs :=
  unfold rget, rset, hget, hset, pcinc, pcset in * .

Ltac simplst := 
  unfoldgs; regfile.simplgs; Heap.simplgs; simplpm.